from binascii import hexlify
from functools import partial
from itertools import chain
from json import dumps, loads
from mimetypes import MimeTypes
from os import stat, urandom
from os.path import exists
from time import sleep
from typing import Any, ByteString, Dict, List, Tuple
from urllib.parse import quote_plus
from urllib.request import Request, urlopen

from .ir.cmd import Cmd
from .ir.django_model import DjangoModel
from .ir.django_property import DjangoProperty
from .loggable import Loggable

charset_map = dict()


class WbDatabase:
    SHRT_MIN = 1
    SHRT_MAX = 1
    INT_MIN = 1
    INT_MAX = 1
    LONG_MIN = 1
    LONG_MAX = 1

    ISOLATION_LEVEL_READ_COMMITED = 1

    _BASE_URL = '_base_url'
    _MEDIAWIKI_VERSION = '_mediawiki_version'
    _INSTANCE_OF = 'instance of'
    _SUBCLASS_OF = 'subclass of'
    _DJANGO_MODEL = 'django model'
    _DJANGO_NAMESPACE = 'django namespace'
    _DJANGO_APPLICATION = 'django application'
    _DJANGO_FIELD = 'django field'
    _DJANGO_SEQUENCE = 'django sequence'
    _DJANGO_NEXT_ID = 'django next id'
    _DJANGO_MODELS = '_django_models'
    _SPARQL_ENDPOINT = '_sparql_endpoint'
    _WIKIBASE_PROPERTIES = '_wikibase_properties'
    _WIKIBASE_CREDENTIALS = '_wikibase_credentials'

    class Error:
        ...

    class DatabaseError(BaseException):
        ...

    class IntegrityError(BaseException):
        ...

    class OperationalError(BaseException):
        ...

    class DataError(BaseException):
        ...

    class InternalError(BaseException):
        ...

    class NotSupportedError(BaseException):
        ...

    class InterfaceError(BaseException):
        ...

    class ProgrammingError(BaseException):
        ...

    @staticmethod
    def get_property_type_for_django_field(field: DjangoProperty) -> str:
        django_field_type = field['property_type']
        if django_field_type == 'CharField':
            return 'string'
        if django_field_type == 'AutoField':
            return 'quantity'
        if django_field_type == 'ForeignKey':
            return 'wikibase-item'
        if django_field_type == 'DecimalField':
            return 'quantity'
        if django_field_type == 'IntegerField':
            return 'quantity'
        if django_field_type == 'DateTimeField':
            return 'time'
        if django_field_type == 'FileField':
            return 'string'
        if django_field_type == 'FloatField':
            return 'quantity'
        raise WbDatabase.InternalError(
            f'Sorry I can\'t recognize type: {django_field_type} for field {field}.')

    @staticmethod
    def connect(charset: str = 'utf8', url: str = '',
                bot_username: str = '', bot_password: str = '',
                instance_of_property_id: int = None,
                subclass_of_property_id: int = None,
                django_model_item_id: int = None,
                wdqs_sparql_endpoint: str = None,
                django_namespace: str = None):
        return WbDatabaseConnection(charset, url,
                                    bot_username, bot_password,
                                    instance_of_property_id,
                                    subclass_of_property_id,
                                    django_model_item_id,
                                    wdqs_sparql_endpoint,
                                    django_namespace)


class WbCredentials(dict):

    def __init__(self, bot_username: str, bot_password: str):
        super().__init__(bot_username=bot_username, bot_password=bot_password)


class WbLink(dict):

    def __init__(self, id: int, entity_type: str, base_url: str):
        super().__init__(id=id, entity_type=entity_type,
                         url=f'{base_url}{self._entity_prefix(entity_type)}{id}')

    @staticmethod
    def _entity_prefix(entity_type: str) -> str:
        if not entity_type:
            return None
        first_letter = str(entity_type).upper()[:1]
        if first_letter == 'I':
            return 'Q'
        if first_letter == 'P':
            return 'P'
        raise WbDatabase.InternalError(
            f'Sorry I can\'t recognize entity type: {entity_type}')

    def get_entity_id(self) -> str:
        return f'{self._entity_prefix(self["entity_type"])}{self["id"]}'


_1X1_JPG = bytearray.fromhex('''
ffd8 ffe0 0010 4a46 4946 0001 0100 0001
0001 0000 ffdb 0043 0003 0202 0202 0203
0202 0203 0303 0304 0604 0404 0404 0806
0605 0609 080a 0a09 0809 090a 0c0f 0c0a
0b0e 0b09 090d 110d 0e0f 1010 1110 0a0c
1213 1210 130f 1010 10ff c000 0b08 0001
0001 0101 1100 ffc4 0014 0001 0000 0000
0000 0000 0000 0000 0000 0009 ffc4 0014
1001 0000 0000 0000 0000 0000 0000 0000
0000 ffda 0008 0101 0000 3f00 2a9f ffd9
''')


class WbApi():

    def __init__(self, url: str, charset: str = 'utf-8', wdqs_sparql_endpoint: str = None):
        super().__init__()
        self.url = url
        self.charset = charset
        self.wdqs_sparql_endpoint = wdqs_sparql_endpoint if wdqs_sparql_endpoint else f'{url}/sparql'
        self.cookies = None
        self.session = None

    def _retry(self, countdown: int, request: Request) -> dict:
        search_result = {}
        for i in range(0, countdown):
            if self.session:
                request.headers['Cookie'] = self.session
            response = urlopen(request)
            search_result = loads(response.read().decode(self.charset))
            if 'error' in search_result and 'code' in search_result['error'] and \
                    (search_result['error']['code'] == 'failed-save' or search_result['error']['code'] == 'no-automatic-entity-id'):
                sleep(1.27 ** i)
                continue
            # Handle session
            if 'Set-Cookie' in response.headers:
                self.cookies = response.headers['Set-Cookie']
                if 'session' in response.headers['Set-Cookie']:
                    self.session = str(self.cookies).split(';')[0]
            return search_result
        raise WbDatabase.InternalError(
            f'Countdown exceeds limit {countdown}. The last search result is {search_result}.')

    def mediawiki_info(self):
        return loads(urlopen(
            f'{self.url}/api.php?action=query&meta=siteinfo&format=json').read().decode(self.charset))

    def search_items(self, query):
        if 'label' in query:
            label_search_string = query['label']
            search_result = loads(urlopen(
                f'{self.url}/api.php?action=wbsearchentities&search={quote_plus(label_search_string)}&language=en&type=item&format=json').read().decode(self.charset))
            return search_result['search']
        raise WbDatabase.InternalError('Only search by label implemented')

    def search_properties(self, query):
        if 'label' in query:
            label_search_string = query['label']
            search_result = loads(urlopen(
                f'{self.url}/api.php?action=wbsearchentities&search={quote_plus(label_search_string)}&language=en&type=property&format=json').read().decode(self.charset))
            return search_result['search']
        raise WbDatabase.InternalError('Only search by label implemented')

    def new_item(self, data):
        retrieve_csrf_token = self._retry(20, Request(
            f'{self.url}/api.php?action=query&meta=tokens&format=json'))
        csrf_token = retrieve_csrf_token['query']['tokens']['csrftoken']
        post_request_body = f'token={quote_plus(csrf_token)}'
        search_result = self._retry(20, Request(
            f'{self.url}/api.php?action=wbeditentity&new=item&data={quote_plus(dumps(data))}&format=json', method='POST', data=post_request_body.encode('utf-8')))
        return search_result['entity']

    def new_property(self, data):
        retrieve_csrf_token = self._retry(20, Request(
            f'{self.url}/api.php?action=query&meta=tokens&format=json'))
        csrf_token = retrieve_csrf_token['query']['tokens']['csrftoken']
        post_request_body = f'token={quote_plus(csrf_token)}'
        search_result = self._retry(20, Request(
            f'{self.url}/api.php?action=wbeditentity&new=property&data={quote_plus(dumps(data))}&format=json', method='POST', data=post_request_body.encode('utf-8')))
        return search_result['entity']

    def get_item_claims(self, numeric_entity_id: int, numeric_property_id: int = None):
        response = urlopen(Request(
            f'{self.url}/api.php?action=wbgetclaims&entity=Q{numeric_entity_id}&format=json' +
            (f'&poperty=P{numeric_property_id}' if numeric_property_id else ''),
            method='GET'))
        search_result = loads(response.read().decode(self.charset))
        return search_result['claims']

    def get_entities(self, wb_links: List[WbLink]) -> List[dict]:
        if not wb_links:
            return []
        entities_ids = '|'.join([wb_link.get_entity_id()
                                for wb_link in wb_links])
        response = urlopen(Request(
            f'{self.url}/api.php?action=wbgetentities&ids={entities_ids}&format=json',
            method='GET'))
        search_result = loads(response.read().decode(self.charset))
        return search_result['entities'].values()

    def new_claim(self, entity_type: str, entity_id: int, property_id: int, value: Any) -> dict:
        entity_id = f'{WbLink._entity_prefix(entity_type)}{entity_id}'
        snak_type = 'value' if value else 'novalue'

        retrieve_csrf_token = self._retry(20, Request(
            f'{self.url}/api.php?action=query&meta=tokens&format=json'))
        csrf_token = retrieve_csrf_token['query']['tokens']['csrftoken']
        post_request_body = f'token={quote_plus(csrf_token)}'
        search_result = self._retry(20, Request(
            f'{self.url}/api.php?action=wbcreateclaim&entity={entity_id}&property=P{property_id}&snaktype={snak_type}&format=json' +
            (f'&value={quote_plus(dumps(value))}' if value else ''), method='POST', data=post_request_body.encode('utf-8')))
        return search_result['claim']

    def get_and_increase_value(self, claim_id: str, increase_step: int) -> int:
        retrieve_csrf_token = self._retry(20, Request(
            f'{self.url}/api.php?action=query&meta=tokens&format=json'))
        csrf_token = retrieve_csrf_token['query']['tokens']['csrftoken']

        response = urlopen(Request(
            f'{self.url}/api.php?action=wbgetclaims&claim={claim_id}&format=json',
            method='GET'))
        search_result = loads(response.read().decode(self.charset))
        for claims in search_result['claims'].values():
            value = claims[0]['mainsnak']['datavalue']['value']
            result = int(value['amount'])
            value['amount'] = result + increase_step
            post_request_body = f'token={quote_plus(csrf_token)}'
            self._retry(20, Request(
                f'{self.url}/api.php?action=wbsetclaimvalue&claim={claim_id}&snaktype=value&&value={quote_plus(dumps(value))}&format=json', method='POST', data=post_request_body.encode('utf-8')))

            return result

        raise WbDatabase.InternalError(
            f'Sorry, I can\'t found claim {claim_id} value')

    def get_claim_value(self, claim_id: str) -> Dict:
        response = urlopen(Request(
            f'{self.url}/api.php?action=wbgetclaims&claim={claim_id}&format=json',
            method='GET'))
        search_result = loads(response.read().decode(self.charset))
        for claims in search_result['claims'].values():
            return claims[0]['mainsnak']['datavalue']['value']

        raise WbDatabase.InternalError(
            f'Sorry, I can\'t found claim {claim_id} value')

    @staticmethod
    def empty_content(guessed_mime_types: Tuple) -> ByteString:
        if guessed_mime_types[0] == 'text/plain':
            return b'\x20'
        if guessed_mime_types[0] == 'image/jpeg':
            return _1X1_JPG
        raise WbDatabase.InternalError(
            f'Sorry, I can\'t generate empty content for the {guessed_mime_types}')

    def read_chunks(self, file_object, chunk_size):
        '''Return the next chunk of the file'''

        while True:
            data = file_object.read(chunk_size)
            if not data:
                break
            yield data

    @staticmethod
    def _build_part(item, sep_boundary):
        key, values = item
        title = '\nContent-Disposition: form-data; name="%s"' % key
        # handle multiple entries for the same name
        if not isinstance(values, list):
            values = [values]
        for value in values:
            if isinstance(value, tuple):
                title += '; filename="%s"' % value[0]
                value = value[1]
            else:
                value = str(value).encode('utf-8', 'surrogateescape')
            yield sep_boundary
            yield title.encode('utf-8', 'surrogateescape')
            yield b"\n\n"
            yield value
            if value and value[-1:] == b'\r':
                yield b'\n'  # write an extra newline (lurve Macs)

    @classmethod
    def encode_multipart_formdata(cls, data):
        '''
        Build up the MIME payload for the POST data
        '''
        boundary = f'--------{hexlify(urandom(16)).decode("ascii")}'
        sep_boundary = b'\n--' + boundary.encode('ascii')
        end_boundary = sep_boundary + b'--'
        end_items = end_boundary, b"\n",
        builder = partial(
            cls._build_part,
            sep_boundary=sep_boundary,
        )
        part_groups = map(builder, data.items())
        parts = chain.from_iterable(part_groups)
        body_items = chain(parts, end_items)
        content_type = 'multipart/form-data; boundary=%s' % boundary
        return b''.join(body_items), content_type

    def login(self, credentials: WbCredentials) -> Dict:
        login_token_response = self._retry(20,
                                           Request(
                                               f'{self.url}/api.php?action=query&meta=tokens&type=login&format=json',
                                               method='GET'))
        data, content_type = self.encode_multipart_formdata({
            'action': 'login',
            'lgname':  credentials['bot_username'],
            'lgpassword': credentials['bot_password'],
            'format': 'json',
            'lgtoken': login_token_response['query']['tokens']['logintoken']
        })
        login_response = self._retry(20,
                                     Request(
                                         f'{self.url}/api.php',
                                         method='POST',
                                         data=data,
                                         headers={'Content-Type': content_type}))
        if not ('login' in login_response) or login_response['login']['result'] != 'Success':
            raise WbDatabase.InternalError(
                f'Sorry, I can\'t login as {credentials["bot_username"]}')
        return login_response

    def upload_file_in_chunks(self, csrf_token: str, file_name: str, file_path: str,
                              file_comment: str = None, chunk_size: int = 65535):
        '''Send multiple post requests to upload a file in chunks using `stash` mode.
        Stash mode is used to build a file up in pieces and then commit it at the end
        '''

        # from mimetypes import MimeTypes
        # guessed_mime_type = MimeTypes().guess_type(file_path)

        file_size = stat(file_path).st_size
        with open(file_path, 'rb') as file_object:
            chunks = self.read_chunks(file_object, chunk_size)
            chunk = next(chunks)

            index = 0
            # Parameters for the first chunk
            params = {
                'action': 'upload',
                'stash': 1,
                'filename': file_name,
                'filesize': file_size,
                'offset': 0,
                'format': 'json',
                'token': csrf_token,
                'ignorewarnings': 1,
                'chunk': ('{}.jpg'.format(index), chunk, 'multipart/form-data')
            }
            data, content_type = self.encode_multipart_formdata(params)
            chunk_upload_result = self._retry(20,
                                              Request(
                                                  f'{self.url}/api.php',
                                                  method='POST',
                                                  data=data,
                                                  headers={'Content-Type': content_type, 'Content-Disposition': '{}.jpg'.format(index)}))
            index += 1

            # Pass the filekey parameter for second and further chunks
            for chunk in chunks:
                params = {
                    'action': 'upload',
                    'stash': 1,
                    'offset': chunk_upload_result['upload']['offset'],
                    'filename': file_name,
                    'filesize': file_size,
                    'filekey': chunk_upload_result['upload']['filekey'],
                    'format': 'json',
                    'token': csrf_token,
                    'ignorewarnings': 1,
                    'chunk': ('{}.jpg'.format(index), chunk, 'multipart/form-data')
                }
                data, content_type = self.encode_multipart_formdata(params)
                chunk_upload_result = self._retry(20,
                                                  Request(
                                                      f'{self.url}/api.php',
                                                      method='POST',
                                                      data=data,
                                                      headers={'Content-Type': content_type}))
                index += 1

            # Final upload using the filekey to commit the upload out of the stash area
            params = {
                'action': 'upload',
                'filename': file_name,
                'filekey': chunk_upload_result['upload']['filekey'],
                'format': 'json',
                'comment': file_comment,
                'token': csrf_token,
            }

            data, content_type = self.encode_multipart_formdata(params)
            chunk_upload_result = self._retry(20,
                                              Request(
                                                  f'{self.url}/api.php',
                                                  method='POST',
                                                  data=data,
                                                  headers={'Content-Type': content_type}))
            return chunk_upload_result

    def upload_file(self, name: str, file_path: str,
                    credentials: WbCredentials) -> Dict:

        self.login(credentials)

        retrieve_csrf_token = self._retry(20, Request(
            f'{self.url}/api.php?action=query&meta=tokens&format=json'))
        csrf_token = retrieve_csrf_token['query']['tokens']['csrftoken']
        # post_request_body = f'token={quote_plus(csrf_token)}'

        upload_result = self.upload_file_in_chunks(csrf_token, name, file_path)

        return upload_result

    def upload_empty_file(self, name: str,
                          credentials: WbCredentials) -> Dict:

        guessed_mime_type = MimeTypes().guess_type(name)

        self.login(credentials)

        retrieve_csrf_token = self._retry(20, Request(
            f'{self.url}/api.php?action=query&meta=tokens&format=json'))
        csrf_token = retrieve_csrf_token['query']['tokens']['csrftoken']
        data, content_type = self.encode_multipart_formdata({
            'action': 'upload',
            'filename': name,
            'format': 'json',
            'token': csrf_token,
            'ignorewarnings': 1,
            'file': (name, self.empty_content(guessed_mime_type), 'multipart/form-data')
        })

        upload_result = self._retry(20,
                                    Request(
                                        f'{self.url}/api.php',
                                        method='POST',
                                        data=data,
                                        headers={'Content-Type': content_type}))

        return upload_result

    def execute_sparql_query(self, sparql_query: str) -> Dict:
        execute_result = self._retry(20,
                                     Request(
                                         f'{self.wdqs_sparql_endpoint}?format=json',
                                         method='POST',
                                         data=sparql_query.encode('utf-8'),
                                         headers={
                                             'Content-Type': 'application/sparql-query',
                                             'Accept': 'application/sparql-results+json'
                                         }))

        return execute_result


class WbCursor(Loggable):

    def __init__(self, django_namespace: str, wikibase_info: dict, api: WbApi):
        super().__init__()
        self.wikibase_info = wikibase_info
        self.django_namespace = django_namespace
        self.api = api
        self.result = []
        self._position = 0

    def close(self):
        print('close')

    def _get_models(self, cmd: Cmd):
        if cmd['cmd'] == 'add_property':
            return (cmd['data']['model'], ) if not cmd['data']['property']['related_models'] else (cmd['data']['model'], cmd['data']['property']['related_models'][0])
        if cmd['cmd'] == 'add_items':
            return (cmd['data']['model'], )
        return None

    def get_or_create_item_if_not_found_by_name(self, item_name: str, data: dict = None) -> Dict:
        entities = self.api.search_items({'label': item_name})
        if len(entities) == 1:
            return entities[0]
        labels_data = {'labels': {
            'en': {'language': 'en', 'value': item_name}}}
        entity_data = labels_data if not data else {**labels_data, **data}
        if len(entities) == 0:
            entity = self.api.new_item(entity_data)
            return entity
        raise WbDatabase.InternalError(
            f'The entity {item_name} has another one (i.e. not unique)')

    def get_or_create_property_if_not_found_by_name(self, property_name: str, data_type_name: str) -> Dict:
        # TODO: add claims if we'll create the new property
        entities = self.api.search_properties({'label': property_name})
        if len(entities) == 1:
            return entities[0]
        if len(entities) == 0:
            entity = self.api.new_property(
                {'labels': {'en': {'language': 'en', 'value': property_name}}, 'datatype': data_type_name})
            return entity
        raise WbDatabase.InternalError(
            f'The property {property_name} has another one (i.e. not unique)')

    def _wb_link(self, snak: dict) -> DjangoProperty:
        return WbLink(
            snak['mainsnak']['datavalue']['value']['numeric-id'],
            snak['mainsnak']['datavalue']['value']['entity-type'],
            self.wikibase_info[WbDatabase._BASE_URL])

    def _general_model_label(self, model: DjangoModel):
        application = model['application']
        return f'{application}{" for " + self.django_namespace if self.django_namespace else ""}'

    def _model_label(self, model: DjangoModel):
        model_name = model['type'].split('.')[-1]
        return f'{model_name} in {self._general_model_label(model)}'

    def _instance_of_model_label(self, model: DjangoModel, id: int):
        model_name = model['type'].split('.')[-1]
        return f'{model_name}:{id} in {self._general_model_label(model)}'

    def _wikibase_property(self, wikibase_property_id: int) -> Dict:
        cached_wikibase_property = self.wikibase_info[WbDatabase._WIKIBASE_PROPERTIES].get(
            wikibase_property_id)
        if cached_wikibase_property:
            return cached_wikibase_property
        for wikibase_property in self.api.get_entities(
                [WbLink(wikibase_property_id, 'property', self.wikibase_info[WbDatabase._BASE_URL])]):
            self.wikibase_info[WbDatabase._WIKIBASE_PROPERTIES][wikibase_property_id] = wikibase_property
            return wikibase_property
        raise WbDatabase.InternalError(
            f'Sorry, I can\'t find wikibase_property  by id {wikibase_property_id}')

    def _claim_item_value(self, numeric_property_id: int, item: Dict = None, numeric_item_id: int = None) -> Dict:
        if item is None and numeric_item_id is None:
            return {
                'mainsnak': {
                    'snaktype': 'novalue',
                    'property': f'P{numeric_property_id}',
                },
                'type': 'statement',
                'rank': 'normal'
            }

        return {
            'mainsnak': {
                'snaktype': 'value',
                'property': f'P{numeric_property_id}',
                'datavalue': {
                            'value': {
                                'entity-type': 'item',
                                'numeric-id': int(item['id'][1:]) if item else numeric_item_id},
                            'type': 'wikibase-entityid'}
            },
            'type': 'statement',
            'rank': 'normal'
        }

    def _claim_string_value(self, numeric_property_id: int, value: str = None) -> Dict:
        if value is None or value == '':
            return {
                'mainsnak': {
                    'snaktype': 'novalue',
                    'property': f'P{numeric_property_id}',
                },
                'type': 'statement',
                'rank': 'normal'
            }

        return {
            'mainsnak': {
                'snaktype': 'value',
                'property': f'P{numeric_property_id}',
                'datavalue': {
                            'value': value,
                            'type': 'string'}
            },
            'type': 'statement',
            'rank': 'normal'
        }

    def _claim_quantity_value(self, numeric_property_id: int, value: None) -> Dict:
        if value is None or value == '':
            return {
                'mainsnak': {
                    'snaktype': 'novalue',
                    'property': f'P{numeric_property_id}',
                },
                'type': 'statement',
                'rank': 'normal'
            }

        return {
            'mainsnak': {
                'snaktype': 'value',
                'property': f'P{numeric_property_id}',
                'datavalue': {
                            'value': {
                                'amount': value,
                                'unit': '1'
                            },
                    'type': 'quantity'}
            },
            'type': 'statement',
            'rank': 'normal'
        }

    def _check_or_create_model(self, model: DjangoModel):

        # Check general model
        general_model_label = self._general_model_label(model)
        if not(general_model_label in self.wikibase_info):
            # get application django model it contains general info
            application_model = self.get_or_create_item_if_not_found_by_name(
                general_model_label,
                data={'claims': [
                    self._claim_item_value(
                        self.wikibase_info[WbDatabase._SUBCLASS_OF]['id'],
                        numeric_item_id=self.wikibase_info[WbDatabase._DJANGO_MODEL]['id']),
                    # self._claim_value(
                    #     self.wikibase_info[WbDatabase._DJANGO_SEQUENCE]['id'],
                    #     numeric_item_id=...),
                    self._claim_string_value(
                        self.wikibase_info[WbDatabase._DJANGO_APPLICATION]['id'],
                        model['application']),
                    self._claim_string_value(
                        self.wikibase_info[WbDatabase._DJANGO_NAMESPACE]['id'],
                        self.django_namespace),
                ]})

            application_model_claims = self.api.get_item_claims(
                int(application_model['id'][1:]))

            django_sequence_property_id = self.wikibase_info[WbDatabase._DJANGO_SEQUENCE]['id']
            if not (f'P{django_sequence_property_id}' in application_model_claims):
                application_model_claims[f'P{django_sequence_property_id}'] = [
                ]
            application_model_sequences = [self._wb_link(property_value) for property_value in application_model_claims[f'P{django_sequence_property_id}']] \
                if f'P{django_sequence_property_id}' in application_model_claims else []

            application_model['claims'] = application_model_claims
            self.wikibase_info[general_model_label] = application_model

        # Check concrete model
        concrete_model_label = self._model_label(model)
        if not(concrete_model_label in self.wikibase_info):

            concrete_model = self.get_or_create_item_if_not_found_by_name(
                concrete_model_label,
                data={'claims': [
                    self._claim_item_value(
                        self.wikibase_info[WbDatabase._SUBCLASS_OF]['id'],
                        self.wikibase_info[general_model_label]),
                    # self._claim_value(
                    #     self.wikibase_info[WbDatabase._DJANGO_SEQUENCE]['id'],
                    #     numeric_item_id=...),
                    self._claim_string_value(
                        self.wikibase_info[WbDatabase._DJANGO_APPLICATION]['id'],
                        model['application']),
                    self._claim_string_value(
                        self.wikibase_info[WbDatabase._DJANGO_NAMESPACE]['id'],
                        self.django_namespace),
                ]})
            concrete_model_id = int(concrete_model['id'][1:])

            django_field_property_id = self.wikibase_info[WbDatabase._DJANGO_FIELD]['id']
            # below is the request existed claims or add a placeholder for them
            concrete_model_claims = self.api.get_item_claims(concrete_model_id)
            if not (f'P{django_field_property_id}' in concrete_model_claims):
                concrete_model_claims[f'P{django_field_property_id}'] = []
            concrete_model_properties = [self._wb_link(property_value) for property_value in concrete_model_claims[f'P{django_field_property_id}']] \
                if f'P{django_field_property_id}' in concrete_model_claims else []

            concrete_model_fields = {
                p['labels']['en']['value'] for p in self.api.get_entities(concrete_model_properties)}

            # Create/Update properties
            for field in model['fields']:
                related_model_label = self._model_label(
                    field['related_models'][0]) if field['related_models'] else None
                related_models = f' to {related_model_label}' if related_model_label else ''
                property_type_name = f'{field["property_type"]}{related_models}'
                property_name = f'{field["property_name"]} type {property_type_name}'

                if property_name in concrete_model_fields:
                    continue

                # Create claim cause missed
                wikibase_property = self.get_or_create_property_if_not_found_by_name(
                    property_name, WbDatabase.get_property_type_for_django_field(field))
                wikibase_property_id = int(wikibase_property['id'][1:])
                self.wikibase_info[WbDatabase._WIKIBASE_PROPERTIES][wikibase_property_id] = wikibase_property

                claim = self.api.new_claim('item', concrete_model_id, django_field_property_id, {
                                           'entity-type': 'property', 'numeric-id': wikibase_property_id})
                concrete_model_claims[f'P{django_field_property_id}'].append(
                    claim)

            # default sequence for the concrete model
            django_next_id_property_id = self.wikibase_info[WbDatabase._DJANGO_NEXT_ID]['id']
            if not (f'P{django_next_id_property_id}' in concrete_model_claims):
                claim = self.api.new_claim('item', concrete_model_id, django_next_id_property_id, {
                                           'amount': 1, 'unit': '1'})
                concrete_model_claims[f'P{django_next_id_property_id}'] = [
                    claim]

            concrete_model['claims'] = concrete_model_claims
            self.wikibase_info[concrete_model_label] = concrete_model
            # Store table link
            self.wikibase_info[WbDatabase._DJANGO_MODELS][model['table_name']] = model

            return concrete_model

        return self.wikibase_info[concrete_model_label]

    def _django_field_name_from_wikibase_property_name(self, wikibase_property_name: str) -> str:
        result = wikibase_property_name.strip()
        return result[:result.index(' ')]

    def _convert_to_snak_and_handle_value(self, wikibase_property_id, wikibase_datatype: str, django_field_value: Any) -> Dict:
        if django_field_value is None or str(django_field_value) == '':
            return {
                'snaktype': 'novalue',
                'property': f'P{wikibase_property_id}'
            }
        result = {
            'snaktype': 'value',
            'property': f'P{wikibase_property_id}',
            'datavalue': {
                'value': None,
                'type': self._replace_unsupported_types(wikibase_datatype)}
        }
        if wikibase_datatype == 'string':
            result['datavalue']['value'] = str(django_field_value)
        elif wikibase_datatype == 'commonsMedia':
            # TODO: upload to the https://commons.wikipedia.org
            result['datavalue']['value'] = django_field_value.name
        elif wikibase_datatype == 'quantity':
            result['datavalue']['value'] = {
                'amount': float(django_field_value),
                'unit': '1'
            }
        else:
            raise WbDatabase.InternalError(
                f'Sorry, I can\'t convert django_field_value {django_field_value} to wikibase_datatype {wikibase_datatype}')

        return result

    def _replace_unsupported_types(self, wikibase_type: str) -> str:
        if wikibase_type == 'commonsMedia':
            return 'string'
        return wikibase_type

    def _claims_make(self, claims: List, value: Dict) -> List:
        result = []
        for claim in claims:
            wikibase_property_id = claim['mainsnak']['datavalue']['value']['numeric-id']
            wikibase_property = self._wikibase_property(wikibase_property_id)
            wikibase_property_name = wikibase_property[
                'label'] if 'label' in wikibase_property else wikibase_property['labels']['en']['value']
            django_field_name = self._django_field_name_from_wikibase_property_name(
                wikibase_property_name)
            if django_field_name in value:
                result.append({
                    'mainsnak': self._convert_to_snak_and_handle_value(wikibase_property_id, wikibase_property['datatype'], value[django_field_name]),
                    'type': 'statement',
                    'rank': 'normal'
                })
        return result

    def _get_autofield_numeric_property_id_or_none(self, model: DjangoModel, concrete_model: Dict) -> int:
        django_field_property_id = self.wikibase_info[WbDatabase._DJANGO_FIELD]['id']
        for field in model['fields']:
            if field['property_type'] == 'AutoField':
                for claim in concrete_model['claims'][f'P{django_field_property_id}']:
                    wikibase_property_id = claim['mainsnak']['datavalue']['value']['numeric-id']
                    wikibase_property = self._wikibase_property(
                        wikibase_property_id)
                    wikibase_property_name = wikibase_property[
                        'label'] if 'label' in wikibase_property else wikibase_property['labels']['en']['value']
                    django_field_name = self._django_field_name_from_wikibase_property_name(
                        wikibase_property_name)
                    if django_field_name == field['attribute_name']:
                        return wikibase_property_id
        return None

    def _get_wikibase_numeric_property_id_or_none(self, model: DjangoModel, property_name: str) -> int:
        concrete_model = self._check_or_create_model(model)
        django_field_property_id = self.wikibase_info[WbDatabase._DJANGO_FIELD]['id']
        for field in model['fields']:
            if field['property_name'] == property_name:
                for claim in concrete_model['claims'][f'P{django_field_property_id}']:
                    wikibase_property_id = claim['mainsnak']['datavalue']['value']['numeric-id']
                    wikibase_property = self._wikibase_property(
                        wikibase_property_id)
                    wikibase_property_name = wikibase_property[
                        'label'] if 'label' in wikibase_property else wikibase_property['labels']['en']['value']
                    django_field_name = self._django_field_name_from_wikibase_property_name(
                        wikibase_property_name)
                    if django_field_name == field['attribute_name']:
                        return wikibase_property_id
        return None

    def _convert_values(self, values_map: dict, keys: list) -> Tuple:
        return tuple([values_map[key]['value'] for key in keys])

    def _tuples(self, vars: list, bindings: list) -> List[Tuple]:
        return [self._convert_values(t, vars) for t in bindings]

    def _add_property(self, cmd: Cmd, params: list):
        self.debug('_add_property %s with %s', cmd, params)

    def _field_indexes(self, cmd: Cmd, params: list):
        self.debug('_field_indexes %s with %s', cmd, params)

    def _create_foreignkey_constraint(self, cmd: Cmd, params: list):
        self.debug('_create_foreignkey_constraint %s with %s', cmd, params)

    def _show_all_models(self, cmd: Cmd, params: list):
        self.debug('_show_all_models %s with %s', cmd, params)

    def _remove_property(self, cmd: Cmd, params: list):
        self.debug('_remove_property %s with %s', cmd, params)

    def _savepoint_create(self, cmd: Cmd, params: list):
        # self.debug('_savepoint_create %s with %s', cmd, params)
        ...

    def _savepoint_rollback(self, cmd: Cmd, params: list):
        # self.debug('_savepoint_rollback %s with %s', cmd, params)
        ...

    def _savepoint_commit(self, cmd: Cmd, params: list):
        # self.debug('_savepoint_commit %s with %s', cmd, params)
        ...

    def _add_items(self, cmd: Cmd, values: list):
        self.debug('_add_items %s with %s', cmd, values)
        model = cmd['data']['model']
        concrete_model = self._check_or_create_model(model)

        instance_of_property_id = self.wikibase_info[WbDatabase._INSTANCE_OF]['id']
        django_field_property_id = self.wikibase_info[WbDatabase._DJANGO_FIELD]['id']
        django_next_id_property_id = self.wikibase_info[WbDatabase._DJANGO_NEXT_ID]['id']

        for value in values:
            next_id = self.api.get_and_increase_value(
                concrete_model['claims'][f'P{django_next_id_property_id}'][0]['id'], 1)

            instance_of_model_label = self._instance_of_model_label(
                model, next_id)
            claims = self._claims_make(
                concrete_model['claims'][f'P{django_field_property_id}'], value)
            claims.append(self._claim_item_value(
                instance_of_property_id, concrete_model))
            autofield_property_id = self._get_autofield_numeric_property_id_or_none(
                model, concrete_model)
            if autofield_property_id:
                claims.append(self._claim_quantity_value(
                    autofield_property_id, next_id))
            self.get_or_create_item_if_not_found_by_name(
                instance_of_model_label,
                data={'claims': claims})

    def _last_insert_id(self, cmd: Cmd, params: list):
        self.debug('_last_insert_id %s with %s', cmd, params)
        model = self.wikibase_info[WbDatabase._DJANGO_MODELS][cmd['data']['table']]
        concrete_model = self._check_or_create_model(model)

        django_next_id_property_id = self.wikibase_info[WbDatabase._DJANGO_NEXT_ID]['id']
        claim_value = self.api.get_claim_value(
            concrete_model['claims'][f'P{django_next_id_property_id}'][0]['id'])
        self.result = [[int(claim_value['amount']) - 1]]
        self._position = 0

    def _set_items(self, cmd: Cmd, values: list):
        self.debug('_set_items %s with %s', cmd, values)
        model = cmd['data']['model']
        concrete_model = self._check_or_create_model(model)

        # TODO: select claims for update

        # instance_of_property_id = self.wikibase_info[WbDatabase._INSTANCE_OF]['id']
        # django_field_property_id = self.wikibase_info[WbDatabase._DJANGO_FIELD]['id']
        # django_next_id_property_id = self.wikibase_info[WbDatabase._DJANGO_NEXT_ID]['id']

        # for value in values:
        #     next_id = self.api.get_and_increase_value(
        #         concrete_model['claims'][f'P{django_next_id_property_id}'][0]['id'], 1)

        #     instance_of_model_label = self._instance_of_model_label(
        #         model, next_id)
        #     claims = self._claims_make(
        #         concrete_model['claims'][f'P{django_field_property_id}'], value)
        #     claims.append(self._claim_item_value(
        #         instance_of_property_id, concrete_model))
        #     self.get_or_create_item_if_not_found_by_name(
        #         instance_of_model_label,
        #         data={'claims': claims})

    def _create_model(self, cmd: Cmd, params: list):
        self.debug('_create_model %s with %s', cmd, params)
        self._check_or_create_model(cmd['data']['model'])

    def _table_exists(self, cmd: Cmd, params: list):
        self.debug('_table_exists %s with %s', cmd, params)

    def _enable_constraints(self, cmd: Cmd, params: list):
        self.debug('_enable_constraints %s with %s', cmd, params)

    def _disable_constraints(self, cmd: Cmd, params: list):
        self.debug('_disable_constraints %s with %s', cmd, params)

    def _create_index(self, cmd: Cmd, params: list):
        self.debug('_create_index %s with %s', cmd, params)

    def _get_constraints(self, cmd: Cmd, params: list):
        self.debug('_get_constraints %s with %s', cmd, params)

    def _select(self, cmd: Cmd, params: list):
        self.debug('_select %s with %s', cmd, params)
        for model in cmd['data']['models']:
            self._check_or_create_model(model)
        # ...
        answer = self.api.execute_sparql_query(cmd['data']['sparql'])
        self.result = self._tuples(
            answer['head']['vars'], answer['results']['bindings'])
        self._position = 0

    def execute(self, cmd, params) -> Any:
        if not isinstance(cmd, Cmd):
            # TODO: transform SQL-92 into commands
            raise WbDatabase.InternalError(f'Wrong command: {cmd}')
        models = self._get_models(cmd)
        if models:
            for model in models:
                self._check_or_create_model(model)
        if cmd['cmd'] == 'add_property':
            return self._add_property(cmd, params)
        if cmd['cmd'] == 'field_indexes':
            return self._field_indexes(cmd, params)
        if cmd['cmd'] == 'create_foreignkey_constraint':
            return self._create_foreignkey_constraint(cmd, params)
        if cmd['cmd'] == 'show_all_models':
            return self._show_all_models(cmd, params)
        if cmd['cmd'] == 'remove_property':
            return self._remove_property(cmd, params)
        if cmd['cmd'] == 'savepoint_create':
            return self._savepoint_create(cmd, params)
        if cmd['cmd'] == 'savepoint_rollback':
            return self._savepoint_rollback(cmd, params)
        if cmd['cmd'] == 'savepoint_commit':
            return self._savepoint_commit(cmd, params)
        if cmd['cmd'] == 'add_items':
            return self._add_items(cmd, params)
        if cmd['cmd'] == 'last_insert_id':
            return self._last_insert_id(cmd, params)
        if cmd['cmd'] == 'set_items':
            return self._set_items(cmd, params)
        if cmd['cmd'] == 'create_model':
            return self._create_model(cmd, params)
        if cmd['cmd'] == 'create_model':
            return self._create_model(cmd, params)
        if cmd['cmd'] == 'table_exists':
            return self._table_exists(cmd, params)
        if cmd['cmd'] == 'enable_constraints':
            return self._enable_constraints(cmd, params)
        if cmd['cmd'] == 'disable_constraints':
            return self._disable_constraints(cmd, params)
        if cmd['cmd'] == 'create_index':
            return self._create_index(cmd, params)
        if cmd['cmd'] == 'get_constraints':
            return self._get_constraints(cmd, params)
        if cmd['cmd'] == 'select':
            return self._select(cmd, params)

        raise WbDatabase.InternalError(
            f'Sorry, but that command {cmd} can\'t execute')

    def fetchall(self):
        '''Fetch rows from the wikibase

        Returns:
            List[Tuple]: List of rows from the wikibase
        '''
        return self.result

    def fetchmany(self, limit: int):
        '''Fetch rows from the wikibase

        Returns:
            List[Tuple]: List of rows from the wikibase
        '''
        self._position += limit
        result = self.result[:limit]
        self.result = self.result[limit:]
        return result

    def fetchone(self):
        '''Fetch single row from the wikibase

        Returns:
            Tuple: the row from the wikibase
        '''
        return self.result[0]

    def rowcount(self):
        return len(self.result)


class WbDatabaseConnection(Loggable):

    def __init__(self, charset: str, url: str,
                 user: str, password: str,
                 instance_of_property_id: int,
                 subclass_of_property_id: int,
                 django_model_item_id: int,
                 wdqs_sparql_endpoint: str,
                 django_namespace: str):
        super().__init__()
        self.charset = charset
        self.api = WbApi(url, charset, wdqs_sparql_endpoint)
        self.user = user
        self.password = password  # TODO: hash instead plain
        self.django_namespace = django_namespace

        if not django_model_item_id:
            # algo: 1. select django model by name 'django model'
            #       2. if not found create the item
            #       3. if found just get the identity
            django_model_item_id = self.create_item_if_not_found_by_name_and_get_id_without_prefix(
                WbDatabase._DJANGO_MODEL)
        django_namespace_property_id = self.create_property_if_not_found_by_name_and_get_id_without_prefix(
            WbDatabase._DJANGO_NAMESPACE, 'string')
        django_application_property_id = self.create_property_if_not_found_by_name_and_get_id_without_prefix(
            WbDatabase._DJANGO_APPLICATION, 'string')
        django_field_property_id = self.create_property_if_not_found_by_name_and_get_id_without_prefix(
            WbDatabase._DJANGO_FIELD, 'wikibase-property')
        django_sequence_property_id = self.create_property_if_not_found_by_name_and_get_id_without_prefix(
            WbDatabase._DJANGO_SEQUENCE, 'wikibase-property')
        django_next_id_property_id = self.create_property_if_not_found_by_name_and_get_id_without_prefix(
            WbDatabase._DJANGO_NEXT_ID, 'quantity')

        mediawiki_info = self.api.mediawiki_info()
        server = mediawiki_info['query']['general']['server']
        # https://avangard.testo.click/api.php?action=query&meta=siteinfo&siprop=extensions&format=json
        self.wikibase_info = {
            WbDatabase._BASE_URL: server,
            WbDatabase._MEDIAWIKI_VERSION: mediawiki_info['query']['general']['generator'],
            WbDatabase._INSTANCE_OF: WbLink(instance_of_property_id, 'property', server),
            WbDatabase._SUBCLASS_OF: WbLink(subclass_of_property_id, 'property', server),
            WbDatabase._DJANGO_MODEL: WbLink(django_model_item_id, 'item', server),
            WbDatabase._DJANGO_NAMESPACE: WbLink(django_namespace_property_id, 'property', server),
            WbDatabase._DJANGO_APPLICATION: WbLink(django_application_property_id, 'property', server),
            WbDatabase._DJANGO_FIELD: WbLink(django_field_property_id, 'property', server),
            WbDatabase._DJANGO_SEQUENCE: WbLink(django_sequence_property_id, 'property', server),
            WbDatabase._DJANGO_NEXT_ID: WbLink(django_next_id_property_id, 'property', server),
            WbDatabase._DJANGO_MODELS: dict(),
            WbDatabase._SPARQL_ENDPOINT: wdqs_sparql_endpoint if wdqs_sparql_endpoint else f'{url}/sparql',
            WbDatabase._WIKIBASE_PROPERTIES: dict(),
            WbDatabase._WIKIBASE_CREDENTIALS: WbCredentials(user, password)
        }
        self.transactions = []
        self._test_django_application_models()

    def prefixes(self) -> List[str]:
        return [
            'PREFIX bd: <http://www.bigdata.com/rdf#>',
            'PREFIX wikibase: <http://wikiba.se/ontology#>',
            'PREFIX e: <%s/entity/>' % (
                self.wikibase_info[WbDatabase._BASE_URL], ),
            'PREFIX pd: <%s/prop/direct/>' % (
                self.wikibase_info[WbDatabase._BASE_URL], )
        ]

    def _test_django_application_models(self):
        test_result = self.api.execute_sparql_query('''
            # Instance of (Document)
            PREFIX bd: <http://www.bigdata.com/rdf#>
            PREFIX wikibase: <http://wikiba.se/ontology#>
            PREFIX e: <%s/entity/>
            PREFIX pd: <%s/prop/direct/>

            SELECT ?item ?itemLabel
            WHERE
            {
                ?item pd:P%s e:Q%s.
                SERVICE wikibase:label { bd:serviceParam wikibase:language "[AUTO_LANGUAGE],en". }
            }

            LIMIT 100
        ''' % (
            self.wikibase_info[WbDatabase._BASE_URL],
            self.wikibase_info[WbDatabase._BASE_URL],
            self.wikibase_info[WbDatabase._SUBCLASS_OF]['id'],
            self.wikibase_info[WbDatabase._DJANGO_MODEL]['id']
        ))
        # TODO: process all django applications
        self.debug(dumps(test_result))

    def create_item_if_not_found_by_name_and_get_id_without_prefix(self, item_name: str) -> int:
        entities = self.api.search_items({'label': item_name})
        if len(entities) == 1:
            return int(entities[0]['id'][1:])
        if len(entities) == 0:
            entity = self.api.new_item(
                {'labels': {'en': {'language': 'en', 'value': item_name}}})
            return int(entity['id'][1:])
        # Exact matching
        for entity in entities:
            if entity['label'] == item_name:
                return int(entity['id'][1:])
        raise WbDatabase.InternalError(
            f'The entity {item_name} has another one (i.e. not unique)')

    def create_property_if_not_found_by_name_and_get_id_without_prefix(self, property_name: str, data_type_name: str):
        entities = self.api.search_properties({'label': property_name})
        if len(entities) == 1:
            return int(entities[0]['id'][1:])
        if len(entities) == 0:
            entity = self.api.new_property(
                {'labels': {'en': {'language': 'en', 'value': property_name}}, 'datatype': data_type_name})
            return int(entity['id'][1:])
        # Exact matching
        for entity in entities:
            if entity['label'] == property_name:
                return int(entity['id'][1:])
        raise WbDatabase.InternalError(
            f'The entity {property_name} has another one (i.e. not unique)')

    def django_model(self, django_table_name: str):
        for table_name, model in self.wikibase_info[WbDatabase._DJANGO_MODELS].items():
            if table_name == django_table_name:
                return model
        raise WbDatabase.InternalError(
            f'Sorry, I can\'t find django model for {django_table_name} in cache {WbDatabase._DJANGO_MODELS}.')

    def expression_instance_of(self, django_table_name: str) -> str:
        concrete_model = self.cursor()._check_or_create_model(
            self.django_model(django_table_name))
        return f'?{django_table_name} pd:P{self.wikibase_info[WbDatabase._INSTANCE_OF]["id"]} e:Q{int(concrete_model["id"][1:])}'

    def expression_has_property(self, django_table_name: str, property_name: str) -> str:
        wikibase_property_id = self.cursor()._get_wikibase_numeric_property_id_or_none(
            self.django_model(django_table_name), property_name)
        return f'?{django_table_name} pd:P{wikibase_property_id} ?{property_name}'

    def db_info(self, key):
        return self.wikibase_info.get(key)

    def cursor(self):
        return WbCursor(self.django_namespace, self.wikibase_info, self.api)

    def rollback(self):
        ...

    def commit(self):
        ...

    def close(self):
        ...

    def trans(self) -> List:
        return self


class TransactionContext:

    def __init__(self, connection: WbDatabaseConnection):
        self.connection = connection

    def __enter__(self):
        # make a database connection and return it
        ...
        return self.connection

    def __exit__(self, exc_type, exc_val, exc_tb):
        # make sure the dbconnection gets closed
        self.connection.close()
        ...



# -*- coding: utf-8 -*-

import django
import os
os.environ.setdefault('DJANGO_SETTINGS_MODULE',
                      'tests.test_main.test_main.settings')
from black_mirror import api
os.environ.setdefault('WIKIBASE_URL', api('wikibase').url)
os.environ.setdefault('SPARQL_ENDPOINT', api('sparql').url)

django.setup()

# End of section for predefined environment values

from django.core.exceptions import ValidationError
from django.test import TestCase

from tests.test_main.model_fields.models import DataModel


# Know issues:
# test_set_and_retrieve (model_fields.test_binaryfield.BinaryFieldTests)
#  File "/usr/lib/python2.7/ctypes/__init__.py", line 65, in create_string_buffer
#    raise TypeError(init)
#  TypeError: F�


class BinaryFieldTests(TestCase):
    binary_data = b'\x00\x46\xFE'

    def test_set_and_retrieve(self):
        data_set = (self.binary_data, memoryview(self.binary_data))
        for bdata in data_set:
            dm = DataModel(data=bdata)
            dm.save()
            dm = DataModel.objects.get(pk=dm.pk)
            self.assertEqual(bytes(dm.data), bytes(bdata))
            # Resave (=update)
            dm.save()
            dm = DataModel.objects.get(pk=dm.pk)
            self.assertEqual(bytes(dm.data), bytes(bdata))
            # Test default value
            self.assertEqual(bytes(dm.short_data), b'\x08')

    def test_max_length(self):
        dm = DataModel(short_data=self.binary_data * 4)
        with self.assertRaises(ValidationError):
            dm.full_clean()



from json.encoder import JSONEncoder
from typing import Any, Tuple

from django.core.exceptions import EmptyResultSet, FieldError
from django.db import DatabaseError, NotSupportedError
from django.db.models.sql import compiler
from django.db.models.sql.where import WhereNode
from django.db.models.lookups import Exact
from django.db.models.expressions import Col
from django.db.models.sql.datastructures import BaseTable, Join
from django.db.transaction import TransactionManagementError

from json import dumps

# The izip_longest was renamed to zip_longest in py3
try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest

from .ir.cmd import Cmd
from .ir.django_model import DjangoModel
from .ir.django_property import DjangoProperty
from .loggable import Loggable


_BUILT_IN_TYPES = {int, float, str, bool}


class SQLCompiler(compiler.SQLCompiler, Loggable):

    _select_data = {
        'models': None,
        'sparql': None
    }

    def resolve_columns(self, row, fields=()):
        # We need to convert values from database to correct django field representation.
        # For instance, if we defined a BooleanField field, django-wikibase do create a
        # smallint field into DB. When retrieving this field value, it's converted to
        # BooleanField again.
        index_start = len(self.query.extra_select)
        values = []
        for value, field in zip_longest(row[index_start:], fields):
            v = self.query.convert_values(
                value, field, connection=self.connection)
            values.append(v)
        return row[:index_start] + tuple(values)

    def compile(self, expression_node) -> Tuple:
        type_of_expression_node = type(expression_node)
        wb_database_connection = self.connection.connection
        if type_of_expression_node is Col:
            if not(expression_node.alias in self.query.alias_refmap):
                self.query.alias_refmap[expression_node.alias] = []
            self.query.alias_refmap[expression_node.alias].append(
                expression_node.field.column)
            # DjangoProperty(expression_node.field, DjangoModel), []
            return f'?{expression_node.field.column}', []
        if type_of_expression_node is BaseTable:
            expressions = [wb_database_connection.expression_instance_of(
                expression_node.table_name)]
            for property_name in self.query.alias_refmap[expression_node.table_name]:
                expressions.append(wb_database_connection.expression_has_property(
                    expression_node.table_name, property_name))
            self.query.alias_refmap[expression_node.table_name].clear()
            return expressions, []
        if type_of_expression_node is WhereNode:
            sql, params = expression_node.as_sql(self, self.connection)
            return sql.replace(' AND ', ' && ').replace(' OR ', ' || '), params
        if type_of_expression_node is Exact:
            lhs_sql, params = self.compile(expression_node.lhs)
            rhs_sql, rhs_params = self.compile(expression_node.rhs)
            if rhs_params:
                params.extend(rhs_params)
            return f'{lhs_sql} = {rhs_sql}', params
        if type_of_expression_node in _BUILT_IN_TYPES:
            return expression_node, None
        if type_of_expression_node is Join:
            sql, params = expression_node.as_sql(self, self.connection)
            return sql.replace(' AND ', ' && ').replace(' OR ', ' || '), params
        raise NotImplementedError(
            f'Sorry, I can\'t perform compile for the node expression {expression_node}')

    def as_sql(self, with_limits=True, with_col_aliases=False):
        refcounts_before = self.query.alias_refcount.copy()
        self.query.alias_refmap = {f: [] for f in refcounts_before}
        models = {self.query.model}
        try:
            if self.query.model:
                self.query.base_table = self.query.model._meta.db_table
            extra_select, order_by, group_by = self.pre_sql_setup()
            for_update_part = None
            # Is a LIMIT/OFFSET clause needed?
            with_limit_offset = with_limits and (
                self.query.high_mark is not None or self.query.low_mark)
            combinator = self.query.combinator
            features = self.connection.features

            result = self.connection.connection.prefixes()

            if combinator:
                if not getattr(features, 'supports_select_{}'.format(combinator)):
                    raise NotSupportedError(
                        '{} is not supported on this database backend.'.format(combinator))
                result, params = self.get_combinator_sql(
                    combinator, self.query.combinator_all)
            else:
                distinct_fields, distinct_params = self.get_distinct()
                # This must come after 'select', 'ordering', and 'distinct'
                # (see docstring of get_from_clause() for details).
                from_, f_params = self.get_from_clause()
                try:
                    sparql_filter, w_params = self.compile(
                        self.where) if self.where is not None else ('', [])
                except EmptyResultSet:
                    if self.elide_empty:
                        raise
                    # Use a predicate that's always False.
                    sparql_filter, w_params = '0 = 1', []
                having, h_params = self.compile(
                    self.having) if self.having is not None else ("", [])
                result.append('SELECT')
                params = []

                if self.query.distinct:
                    distinct_result, distinct_params = self.connection.ops.distinct_sql(
                        distinct_fields,
                        distinct_params,
                    )
                    result += distinct_result
                    params += distinct_params

                out_cols = []
                col_idx = 1
                for col, (s_sql, s_params), alias in self.select + extra_select:
                    # if alias:
                    #     s_sql = '%s AS %s' % (
                    #         s_sql, self.connection.ops.quote_name(alias))
                    # elif with_col_aliases:
                    #     s_sql = '%s AS %s' % (
                    #         s_sql,
                    #         self.connection.ops.quote_name('col%d' % col_idx),
                    #     )
                    #     col_idx += 1
                    params.extend(s_params)
                    # out_cols.append(s_sql)

                    # update models
                    models.add(col.field.model)
                    out_cols.append(s_sql)

                # result += [', '.join(out_cols), 'FROM', *from_]
                result += [' '.join(out_cols), 'WHERE {', '\n . '.join(*from_)]
                params.extend(f_params)

                if self.query.select_for_update and self.connection.features.has_select_for_update:
                    if self.connection.get_autocommit():
                        raise TransactionManagementError(
                            'select_for_update cannot be used outside of a transaction.')

                    if with_limit_offset and not self.connection.features.supports_select_for_update_with_limit:
                        raise NotSupportedError(
                            'LIMIT/OFFSET is not supported with '
                            'select_for_update on this database backend.'
                        )
                    nowait = self.query.select_for_update_nowait
                    skip_locked = self.query.select_for_update_skip_locked
                    of = self.query.select_for_update_of
                    no_key = self.query.select_for_no_key_update
                    # If it's a NOWAIT/SKIP LOCKED/OF/NO KEY query but the
                    # backend doesn't support it, raise NotSupportedError to
                    # prevent a possible deadlock.
                    if nowait and not self.connection.features.has_select_for_update_nowait:
                        raise NotSupportedError(
                            'NOWAIT is not supported on this database backend.')
                    elif skip_locked and not self.connection.features.has_select_for_update_skip_locked:
                        raise NotSupportedError(
                            'SKIP LOCKED is not supported on this database backend.')
                    elif of and not self.connection.features.has_select_for_update_of:
                        raise NotSupportedError(
                            'FOR UPDATE OF is not supported on this database backend.')
                    elif no_key and not self.connection.features.has_select_for_no_key_update:
                        raise NotSupportedError(
                            'FOR NO KEY UPDATE is not supported on this '
                            'database backend.'
                        )
                    for_update_part = self.connection.ops.for_update_sql(
                        nowait=nowait,
                        skip_locked=skip_locked,
                        of=self.get_select_for_update_of_arguments(),
                        no_key=no_key,
                    )

                if for_update_part and self.connection.features.for_update_after_from:
                    result.append(for_update_part)

                if sparql_filter:
                    result.append(' . FILTER (%s)' % sparql_filter)
                    params.extend(w_params)

                result.append('}')

                grouping = []
                for g_sql, g_params in group_by:
                    grouping.append(g_sql)
                    params.extend(g_params)
                if grouping:
                    if distinct_fields:
                        raise NotImplementedError(
                            'annotate() + distinct(fields) is not implemented.')
                    order_by = order_by or self.connection.ops.force_no_ordering()
                    result.append('GROUP BY %s' % ', '.join(grouping))
                    if self._meta_ordering:
                        order_by = None
                if having:
                    result.append('HAVING %s' % having)
                    params.extend(h_params)

            if self.query.explain_info:
                result.insert(0, self.connection.ops.explain_query_prefix(
                    self.query.explain_info.format,
                    **self.query.explain_info.options
                ))

            if order_by:
                ordering = []
                for _, (o_sql, o_params, _) in order_by:
                    ordering.append(o_sql)
                    params.extend(o_params)
                result.append('ORDER BY %s' % ', '.join(ordering))

            if with_limit_offset:
                result.append(self.connection.ops.limit_offset_sql(
                    self.query.low_mark, self.query.high_mark))

            if for_update_part and not self.connection.features.for_update_after_from:
                result.append(for_update_part)

            if self.query.subquery and extra_select:
                # If the query is used as a subquery, the extra selects would
                # result in more columns than the left-hand side expression is
                # expecting. This can happen when a subquery uses a combination
                # of order_by() and distinct(), forcing the ordering expressions
                # to be selected as well. Wrap the query in another subquery
                # to exclude extraneous selects.
                sub_selects = []
                sub_params = []
                for index, (select, _, alias) in enumerate(self.select, start=1):
                    if not alias and with_col_aliases:
                        alias = 'col%d' % index
                    if alias:
                        sub_selects.append("%s.%s" % (
                            self.connection.ops.quote_name('subquery'),
                            self.connection.ops.quote_name(alias),
                        ))
                    else:
                        select_clone = select.relabeled_clone(
                            {select.alias: 'subquery'})
                        subselect, subparams = select_clone.as_sql(
                            self, self.connection)
                        sub_selects.append(subselect)
                        sub_params.extend(subparams)
                return 'SELECT %s FROM (%s) subquery' % (
                    ', '.join(sub_selects),
                    ' '.join(result),
                ), tuple(sub_params + params)

        finally:
            # Finally do cleanup - get rid of the joins we created above.
            self.query.reset_refcounts(refcounts_before)

        cmd_data = {**self._select_data, **{
            'models': [DjangoModel(model) for model in models],
            'sparql': '\n'.join(result)
        }}
        return Cmd('select', cmd_data), params


class SQLInsertCompiler(compiler.SQLInsertCompiler, Loggable):

    _add_items_data = {
        'model': None,
        'fields': None,
    }

    def as_sql(self, with_limits=True, with_col_aliases=False):
        # sql, params = super(compiler.SQLInsertCompiler, self).as_sql(
        #     with_limits=with_limits, with_col_aliases=with_col_aliases)
        self.debug('SQL insert compiler, as_sql, with limits %s, with col aliases %s',
                   with_limits, with_col_aliases)
        cmd_data = {**self._add_items_data, **{
            'model': DjangoModel(self.query.fields[0].model),
            'fields': [DjangoProperty(f, DjangoModel) for f in self.query.fields]
        }}

        return [(Cmd('add_items', cmd_data),
                 [{f['property_name']: getattr(o, f['attribute_name']) for f in cmd_data['fields']} for o in self.query.objs])]


class SQLDeleteCompiler(compiler.SQLDeleteCompiler, Loggable):

    _remove_items_data = {
        'sparql': None
    }

    def as_sql(self, with_limits=True, with_col_aliases=False):
        sql, params = super(compiler.SQLDeleteCompiler, self).as_sql(
            with_limits=False, with_col_aliases=with_col_aliases)
        self.debug('SQL delete compiler, as_sql, with limits %s, with col aliases %s',
                   with_limits, with_col_aliases)
        cmd_data = {**self._remove_items_data, **{'sparql': sql}}
        return [(Cmd('remove_items', cmd_data), (params))]


class SQLUpdateCompiler(compiler.SQLUpdateCompiler, Loggable):

    _set_items_data = {
        'model': None,
        'values': None,
        'where': None
    }

    def as_sql(self, with_limits=True, with_col_aliases=False):
        # sql, params = super(compiler.SQLUpdateCompiler, self).as_sql(
        #     with_limits=False, with_col_aliases=with_col_aliases)
        self.debug('SQL update compiler, as_sql, with limits %s, with col aliases %s',
                   with_limits, with_col_aliases)
        self.pre_sql_setup()
        cmd_data = {**self._set_items_data, **{
            'model': DjangoModel(self.query.model),
            'values': [DjangoProperty(f, DjangoModel) for f, _, _ in self.query.values],
            'where': self.query.where.as_sql(self, self.connection)
        }}

        return Cmd('set_items', cmd_data), [{f['property_name']: getattr(v, f['attribute_name']) for f in cmd_data['values']} for _, _, v in self.query.values]


class SQLAggregateCompiler(compiler.SQLAggregateCompiler, Loggable):

    _agg_items_data = {
        'sparql': None
    }

    def as_sql(self, with_limits=True, with_col_aliases=False):
        sql, params = super(compiler.SQLAggregateCompiler, self).as_sql(
            with_limits=False, with_col_aliases=with_col_aliases)
        self.debug('SQL aggregate compiler, as_sql, with limits %s, with col aliases %s',
                   with_limits, with_col_aliases)
        cmd_data = {**self._agg_items_data, **{'sparql': sql}}
        return [(Cmd('agg_items', cmd_data), (params))]



"""
    upload_file_in_chunks.py

    MediaWiki API Demos
    Demo of `Upload` module: Step-by-step process to upload a file in chunks

    MIT license
"""
# https://www.mediawiki.org/wiki/API:Upload

import os
import requests
from .one_pixel_jpeg import make_one_pixel_jpeg

S = requests.Session()
URL = "http://localhost:8371/api.php"

# File path of the image to be uploaded
FILE_PATH = '/tmp/f.jpg'
make_one_pixel_jpeg(FILE_PATH)
FILE = open(FILE_PATH, 'rb')
FILE_SIZE = os.stat(FILE_PATH).st_size

# Bot credentials
BOT_USERNAME = 'enter_a_bot_username'
BOT_PASSWORD = 'enter_a_bot_password'

def read_chunks(file_object, chunk_size=5000):
    """Return the next chunk of the file"""

    while True:
        data = file_object.read(chunk_size)
        if not data:
            break
        yield data

def fetch_login_token():
    """Retrieve a login token"""

    params = {
        "action": "query",
        "meta": "tokens",
        "type": "login",
        "format": "json"
    }
    res = S.get(url=URL, params=params)
    data = res.json()
    return data["query"]["tokens"]["logintoken"]

def user_login(login_token, username, password):
    """Send a post request to login
    Use of main account for login is not supported. Obtain credentials via
    Special:BotPasswords (https://www.mediawiki.org/wiki/Special:BotPasswords)
    for lgname & lgpassword """

    params = {
        "action": "login",
        "lgname": username,
        "lgpassword": password,
        "format": "json",
        "lgtoken": login_token
    }
    S.post(URL, data=params)

def fetch_csrf_token():
    """While logged in, retrieve a CSRF token"""

    params = {
        "action": "query",
        "meta":"tokens",
        "format":"json"
    }
    res = S.get(url=URL, params=params)
    data = res.json()
    return data["query"]["tokens"]["csrftoken"]

def upload_file_in_chunks(csrf_token):
    """Send multiple post requests to upload a file in chunks using `stash` mode.
    Stash mode is used to build a file up in pieces and then commit it at the end
    """

    chunks = read_chunks(FILE)
    chunk = next(chunks)

    # Parameters for the first chunk
    params = {
        "action": "upload",
        "stash": 1,
        "filename": "chunk_test.jpg",
        "filesize": FILE_SIZE,
        "offset": 0,
        "format": "json",
        "token": csrf_token,
        "ignorewarnings": 1
    }
    index = 0
    file = {'chunk':('{}.jpg'.format(index), chunk, 'multipart/form-data')}
    index += 1
    res = S.post(URL, files=file, data=params)
    data = res.json()

    # Pass the filekey parameter for second and further chunks
    for chunk in chunks:
        params = {
            "action": "upload",
            "stash": 1,
            "offset": data["upload"]["offset"],
            "filename": "chunk_test.jpg",
            "filesize": FILE_SIZE,
            "filekey": data["upload"]["filekey"],
            "format": "json",
            "token": csrf_token,
            "ignorewarnings": 1
        }
        file = {'chunk':('{}.jpg'.format(index), chunk, 'multipart/form-data')}
        index += 1
        res = S.post(URL, files=file, data=params)
        data = res.json()

    # Final upload using the filekey to commit the upload out of the stash area
    params = {
        "action": "upload",
        "filename": "chunk_test.jpg",
        "filekey": data["upload"]["filekey"],
        "format": "json",
        "comment": "Upload Testing",
        "token": csrf_token,
    }
    res = S.post(URL, data=params)
    data = res.json()

def main():
    """ Four-step process to uploading a file in chunks"""

    login_token = fetch_login_token() # Step 1: Fetch login token
    user_login(login_token, BOT_USERNAME, BOT_PASSWORD) # Step 2: Login
    csrf_token = fetch_csrf_token() # Step 3: Fetch CSRF token
    upload_file_in_chunks(csrf_token) # Step 3: Upload a file in chunks

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import django
import os
os.environ.setdefault('DJANGO_SETTINGS_MODULE',
                      'tests.test_main.test_main.settings')
from black_mirror import api
os.environ.setdefault('WIKIBASE_URL', api('wikibase').url)
os.environ.setdefault('SPARQL_ENDPOINT', api('sparql').url)

django.setup()

# End of section for predefined environment values

import datetime
import json

from django import forms
from django.core import exceptions, serializers
from django.db import models
from django.test import SimpleTestCase, TestCase

from tests.test_main.model_fields.models import DurationModel, NullDurationModel


class TestSaveLoad(TestCase):

    def test_simple_roundtrip(self):
        duration = datetime.timedelta(days=123, seconds=123, microseconds=123)
        DurationModel.objects.create(field=duration)
        loaded = DurationModel.objects.get()
        self.assertEqual(loaded.field, duration)

    def test_create_empty(self):
        NullDurationModel.objects.create()
        loaded = NullDurationModel.objects.get()
        self.assertIsNone(loaded.field)

    def test_fractional_seconds(self):
        value = datetime.timedelta(seconds=2.05)
        d = DurationModel.objects.create(field=value)
        d.refresh_from_db()
        self.assertEqual(d.field, value)


class TestQuerying(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.objs = [
            DurationModel.objects.create(field=datetime.timedelta(days=1)),
            DurationModel.objects.create(field=datetime.timedelta(seconds=1)),
            DurationModel.objects.create(field=datetime.timedelta(seconds=-1)),
        ]

    def test_exact(self):
        self.assertSequenceEqual(
            DurationModel.objects.filter(field=datetime.timedelta(days=1)),
            [self.objs[0]]
        )

    def test_gt(self):
        self.assertSequenceEqual(
            DurationModel.objects.filter(field__gt=datetime.timedelta(days=0)),
            [self.objs[0], self.objs[1]]
        )


class TestSerialization(SimpleTestCase):
    test_data = '[{"fields": {"field": "1 01:00:00"}, "model": "model_fields.durationmodel", "pk": null}]'

    def test_dumping(self):
        instance = DurationModel(field=datetime.timedelta(days=1, hours=1))
        data = serializers.serialize('json', [instance])
        self.assertEqual(json.loads(data), json.loads(self.test_data))

    def test_loading(self):
        instance = list(serializers.deserialize('json', self.test_data))[0].object
        self.assertEqual(instance.field, datetime.timedelta(days=1, hours=1))


class TestValidation(SimpleTestCase):

    def test_invalid_string(self):
        field = models.DurationField()
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean('not a datetime', None)
        self.assertEqual(cm.exception.code, 'invalid')
        self.assertEqual(
            cm.exception.message % cm.exception.params,
            "'not a datetime' value has an invalid format. "
            "It must be in [DD] [HH:[MM:]]ss[.uuuuuu] format."
        )


class TestFormField(SimpleTestCase):
    # Tests for forms.DurationField are in the forms_tests app.

    def test_formfield(self):
        field = models.DurationField()
        self.assertIsInstance(field.formfield(), forms.DurationField)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import django
import os
os.environ.setdefault('DJANGO_SETTINGS_MODULE',
                      'tests.test_main.test_main.settings')
from black_mirror import api
os.environ.setdefault('WIKIBASE_URL', api('wikibase').url)
os.environ.setdefault('SPARQL_ENDPOINT', api('sparql').url)

django.setup()

# End of section for predefined environment values

from unittest import skipIf

from django.core.exceptions import ValidationError
from django.db import connection, models
from django.test import SimpleTestCase, TestCase
from django.utils.functional import lazy

from tests.test_main.model_fields.models import Post


class CharFieldTests(TestCase):

    def test_max_length_passed_to_formfield(self):
        """
        CharField passes its max_length attribute to form fields created using
        the formfield() method.
        """
        cf1 = models.CharField()
        cf2 = models.CharField(max_length=1234)
        self.assertIsNone(cf1.formfield().max_length)
        self.assertEqual(1234, cf2.formfield().max_length)

    def test_lookup_integer_in_charfield(self):
        self.assertEqual(Post.objects.filter(title=9).count(), 0)

    @skipIf(connection.vendor == 'mysql', 'Running on MySQL requires utf8mb4 encoding (#18392)')
    def test_emoji(self):
        p = Post.objects.create(title='Smile 😀', body='Whatever.')
        p.refresh_from_db()
        self.assertEqual(p.title, 'Smile 😀')


class ValidationTests(SimpleTestCase):

    def test_charfield_raises_error_on_empty_string(self):
        f = models.CharField()
        with self.assertRaises(ValidationError):
            f.clean('', None)

    def test_charfield_cleans_empty_string_when_blank_true(self):
        f = models.CharField(blank=True)
        self.assertEqual('', f.clean('', None))

    def test_charfield_with_choices_cleans_valid_choice(self):
        f = models.CharField(max_length=1, choices=[('a', 'A'), ('b', 'B')])
        self.assertEqual('a', f.clean('a', None))

    def test_charfield_with_choices_raises_error_on_invalid_choice(self):
        f = models.CharField(choices=[('a', 'A'), ('b', 'B')])
        with self.assertRaises(ValidationError):
            f.clean('not a', None)

    def test_charfield_get_choices_with_blank_defined(self):
        f = models.CharField(choices=[('', '<><>'), ('a', 'A')])
        self.assertEqual(f.get_choices(True), [('', '<><>'), ('a', 'A')])

    def test_charfield_get_choices_doesnt_evaluate_lazy_strings(self):
        # Regression test for #23098
        # Will raise ZeroDivisionError if lazy is evaluated
        lazy_func = lazy(lambda x: 0 / 0, int)
        f = models.CharField(choices=[(lazy_func('group'), (('a', 'A'), ('b', 'B')))])
        self.assertEqual(f.get_choices(True)[0], ('', '---------'))

    def test_charfield_raises_error_on_empty_input(self):
        f = models.CharField(null=False)
        with self.assertRaises(ValidationError):
            f.clean(None, None)



# -*- coding: utf-8 -*-

import django
import os
os.environ.setdefault('DJANGO_SETTINGS_MODULE',
                      'tests.test_main.test_main.settings')
from black_mirror import api
os.environ.setdefault('WIKIBASE_URL', api('wikibase').url)
os.environ.setdefault('SPARQL_ENDPOINT', api('sparql').url)

django.setup()

# End of section for predefined environment values

from decimal import Decimal

from django.apps import apps
from django.core import checks
from django.db import models
from django.test import TestCase, skipIfDBFeature
from django.test.utils import isolate_apps
from six import text_type

from tests.test_main.model_fields.models import Bar, FkToChar, Foo, PrimaryKeyCharModel


class ForeignKeyTests(TestCase):

    def test_callable_default(self):
        """A lazy callable may be used for ForeignKey.default."""
        a = Foo.objects.create(id=1, a='abc', d=Decimal('12.34'))
        b = Bar.objects.create(b='bcd')
        self.assertEqual(b.a, a)

    @skipIfDBFeature('interprets_empty_strings_as_nulls')
    def test_empty_string_fk(self):
        """
        Empty strings foreign key values don't get converted to None (#19299).
        """
        char_model_empty = PrimaryKeyCharModel.objects.create(string='')
        fk_model_empty = FkToChar.objects.create(out=char_model_empty)
        fk_model_empty = FkToChar.objects.select_related('out').get(id=fk_model_empty.pk)
        self.assertEqual(fk_model_empty.out, char_model_empty)

    @isolate_apps('model_fields')
    def test_warning_when_unique_true_on_fk(self):
        class Foo(models.Model):
            pass

        class FKUniqueTrue(models.Model):
            fk_field = models.ForeignKey(Foo, models.CASCADE, unique=True)

        model = FKUniqueTrue()
        expected_warnings = [
            checks.Warning(
                'Setting unique=True on a ForeignKey has the same effect as using a OneToOneField.',
                hint='ForeignKey(unique=True) is usually better served by a OneToOneField.',
                obj=FKUniqueTrue.fk_field.field,
                id='fields.W342',
            )
        ]
        warnings = model.check()
        self.assertEqual(warnings, expected_warnings)

    def test_related_name_converted_to_text(self):
        rel_name = Bar._meta.get_field('a').remote_field.related_name
        self.assertIsInstance(rel_name, text_type)

    def test_abstract_model_pending_operations(self):
        """
        Foreign key fields declared on abstract models should not add lazy
        relations to resolve relationship declared as string (#24215).
        """
        pending_ops_before = list(apps._pending_operations.items())

        class AbstractForeignKeyModel(models.Model):
            fk = models.ForeignKey('missing.FK', models.CASCADE)

            class Meta:
                abstract = True

        self.assertIs(AbstractForeignKeyModel._meta.apps, apps)
        self.assertEqual(
            pending_ops_before,
            list(apps._pending_operations.items()),
            'Pending lookup added for a foreign key on an abstract model'
        )

    @isolate_apps('model_fields', 'model_fields.tests')
    def test_abstract_model_app_relative_foreign_key(self):
        class AbstractReferent(models.Model):
            reference = models.ForeignKey('Referred', on_delete=models.CASCADE)

            class Meta:
                app_label = 'model_fields'
                abstract = True

        def assert_app_model_resolved(label):
            class Referred(models.Model):
                class Meta:
                    app_label = label

            class ConcreteReferent(AbstractReferent):
                class Meta:
                    app_label = label

            self.assertEqual(ConcreteReferent._meta.get_field('reference').related_model, Referred)

        assert_app_model_resolved('model_fields')
        assert_app_model_resolved('tests')



"""
    upload_file_from_url.py

    MediaWiki API Demos
    Demo of `Upload` module: Post request to upload a file from a URL

    MIT license
"""
# https://www.mediawiki.org/wiki/API:Upload

import requests

S = requests.Session()
URL = "http://localhost:8371/api.php"

# Step 1: Retrieve a login token
PARAMS_1 = {
    "action": "query",
    "meta": "tokens",
    "type": "login",
    "format": "json"
}

R = S.get(url=URL, params=PARAMS_1)
DATA = R.json()

LOGIN_TOKEN = DATA["query"]["tokens"]["logintoken"]

# Step 2: Send a post request to login. Use of main account for login is not
# supported. Obtain credentials via Special:BotPasswords
# (https://www.mediawiki.org/wiki/Special:BotPasswords) for lgname & lgpassword
PARAMS_2 = {
    "action": "login",
    "lgname": "bot_username",
    "lgpassword": "bot_password",
    "format": "json",
    "lgtoken": LOGIN_TOKEN
}

R = S.post(URL, data=PARAMS_2)

# Step 3: While logged in, retrieve a CSRF token
PARAMS_3 = {
    "action": "query",
    "meta":"tokens",
    "format":"json"
}

R = S.get(url=URL, params=PARAMS_3)
DATA = R.json()

CSRF_TOKEN = DATA["query"]["tokens"]["csrftoken"]

# Step 4: Post request to upload a file from a URL
PARAMS_4 = {
    "action": "upload",
    "filename": "new_image.jpg",
    "url": "https://farm9.staticflickr.com/8213/8300206113_374c017fc5.jpg",
    "format": "json",
    "token": CSRF_TOKEN,
    "ignorewarnings": 1
}

R = S.post(URL, data=PARAMS_4)
DATA = R.json()
print(DATA)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import django
import os
os.environ.setdefault('DJANGO_SETTINGS_MODULE',
                      'tests.test_main.test_main.settings')
from black_mirror import api
os.environ.setdefault('WIKIBASE_URL', api('wikibase').url)
os.environ.setdefault('SPARQL_ENDPOINT', api('sparql').url)

django.setup()

# End of section for predefined environment values

from decimal import Decimal
from unittest import skipUnless

from django.core import validators
from django.core.exceptions import ValidationError
from django.db import models
from django.test import TestCase

from tests.test_main.model_fields.models import BigD, Foo


class DecimalFieldTests(TestCase):

    def test_to_python(self):
        f = models.DecimalField(max_digits=4, decimal_places=2)
        self.assertEqual(f.to_python(3), Decimal('3'))
        self.assertEqual(f.to_python('3.14'), Decimal('3.14'))
        with self.assertRaises(ValidationError):
            f.to_python('abc')

    def test_default(self):
        f = models.DecimalField(default=Decimal('0.00'))
        self.assertEqual(f.get_default(), Decimal('0.00'))

    @skipUnless(False, "DecimalField hasn't _format() method")
    def test_format(self):
        f = models.DecimalField(max_digits=5, decimal_places=1)
        self.assertEqual(f._format(f.to_python(2)), '2.0')
        self.assertEqual(f._format(f.to_python('2.6')), '2.6')
        self.assertIsNone(f._format(None))

    def test_get_prep_value(self):
        f = models.DecimalField(max_digits=5, decimal_places=1)
        self.assertIsNone(f.get_prep_value(None))
        self.assertEqual(f.get_prep_value('2.4'), Decimal('2.4'))

    def test_filter_with_strings(self):
        """
        Should be able to filter decimal fields using strings (#8023).
        """
        foo = Foo.objects.create(a='abc', d=Decimal('12.34'))
        self.assertEqual(list(Foo.objects.filter(d='12.34')), [foo])

    def test_save_without_float_conversion(self):
        """
        Ensure decimals don't go through a corrupting float conversion during
        save (#5079).
        """
        bd = BigD(d='12.9')
        bd.save()
        bd = BigD.objects.get(pk=bd.pk)
        self.assertEqual(bd.d, Decimal('12.9'))

    def test_lookup_really_big_value(self):
        """
        Really big values can be used in a filter statement.
        """
        # This should not crash.
        Foo.objects.filter(d__gte=100000000000)

    def test_max_digits_validation(self):
        field = models.DecimalField(max_digits=2)
        expected_message = validators.DecimalValidator.messages['max_digits'] % {'max': 2}
        with self.assertRaisesMessage(ValidationError, expected_message):
            field.clean(100, None)

    def test_max_decimal_places_validation(self):
        field = models.DecimalField(decimal_places=1)
        expected_message = validators.DecimalValidator.messages['max_decimal_places'] % {'max': 1}
        with self.assertRaisesMessage(ValidationError, expected_message):
            field.clean(Decimal('0.99'), None)

    def test_max_whole_digits_validation(self):
        field = models.DecimalField(max_digits=3, decimal_places=1)
        expected_message = validators.DecimalValidator.messages['max_whole_digits'] % {'max': 2}
        with self.assertRaisesMessage(ValidationError, expected_message):
            field.clean(Decimal('999'), None)



from logging import getLogger, root, DEBUG, Logger, StreamHandler


class Loggable:

    logger: Logger = None

    def _init_logger(self):
        self.logger = getLogger(f'{__name__}.{self.__class__.__name__}')

    def __init__(self):
        self._init_logger()

    def debug(self, *args, **kwargs):
        if root.level <= DEBUG:
            if not self.logger:
                self._init_logger()
            self.logger.debug(*args, **kwargs)
